package spire.math

import spire.algebra._

import scala.reflect.ClassTag
import scala.collection.mutable
import scala.{specialized => spec}

trait Range[@spec(Int, Long) A] {
  def start: A
  def end: A
  def step: A
  def length: Long
  def foreach(f: A => Unit): Unit
  def toList: List[A] = {
    val buf = mutable.ListBuffer.empty[A]
    foreach { a => buf.append(a) }
    buf.toList
  }
  def toArray(implicit ct: ClassTag[A]): Array[A] = {
    if (length > Int.MaxValue)
      sys.error("length(%s) exceeds Int.MaxValue" format length)
    val arr = new Array[A](length.toInt)
    var i = 0
    foreach { a => arr(i) = a; i += 1 }
    arr
  }
  def foldLeft[@spec B](init: B)(f: (B, A) => B): B = {
    var b = init
    foreach(a => b = f(b, a))
    b
  }
  def contains(a: A): Boolean
}

private[math] class EmptyRange[A](implicit m: Monoid[A]) extends Range[A] {
  def start: A = m.id
  def end: A = m.id
  def step: A = m.id
  def length: Long = 0L
  def foreach(f: A => Unit): Unit = ()
  def contains(a: A) = false
}

private[math] case class IntRange(start: Int, end: Int, step: Int) extends Range[Int] {
  private def initLength(): Long = {
    if (end < start) sys.error("end=%s < start=%s" format (end, start))
    if (step < 0) sys.error("negative step=%s" format step)

    (end.toLong - start.toLong) / step
  }

  final val length: Long = initLength()

  def foreach(f: Int => Unit) {
    val limit = end - step
    var i = start
    while (i <= limit) { f(i); i += step }
    f(i)
  }
  def contains(a: Int) = a >= start && a <= end && (a - start) % step == 0
}

private[math] case class LongRange(start: Long, end: Long, step: Long) extends Range[Long] {
  private def initLength(): Long = {
    if (end < start) sys.error("end=%s < start=%s" format (end, start))
    if (step < 0L) sys.error("negative step=%s" format step)
    val q = (end / step) - (start / step)
    val m = (end % step) - (start % step)
    val len = if (m < 0L) q - 1 else q
    if (len < 0L) sys.error("length(%s) overflows" format len)
    len
  }

  final val length: Long = initLength()

  def foreach(f: Long => Unit) {
    val limit = end - step
    var i = start
    while (i <= limit) { f(i); i += step }
    f(i)
  }
  def contains(a: Long) = a >= start && a <= end && (a - start) % step == 0L
}

private[math] case class RationalRange(start: Rational, end: Rational, step: Rational) extends Range[Rational] {
  private def initLength(): Long = {
    if (end < start) sys.error("end=%s < start=%s" format (end, start))
    if (step < Rational.zero) sys.error("negative step=%s" format step)
    val q = (end quot step) - (start quot step)
    val m = (end % step) - (start % step)
    val len = if (m < Rational.zero) q - 1 else q
    if (len > Long.MaxValue) sys.error("length(%s) overflows" format len)
    len.toLong
  }

  final val length: Long = initLength()

  def foreach(f: Rational => Unit) {
    val limit = end - step
    var i = start
    while (i <= limit) { f(i); i += step }
    f(i)
  }
  def contains(a: Rational) =
    a >= start && a <= end && (a - start) % step == Rational.zero
}

object Range {
  def apply(start: Int, end: Int, step: Int): Range[Int] =
    IntRange(start, end, step)
  def apply(start: Long, end: Long, step: Long): Range[Long] =
    LongRange(start, end, step)
  def apply(start: Rational, end: Rational, step: Rational): Range[Rational] =
    RationalRange(start, end, step)
}
